@@ -22,6 +22,85 @@
 #include "linux_version.h"
 #include "mountP.h"
 
+/*
+ * Kernel supports only one MS_PROPAGATION flag change by one mount(2) syscall,
+ * to bypass this restriction we call mount(2) per flag. It's really not a perfect
+ * solution, but it's the same like to execute multiple mount(8) commands.
+ *
+ * We use cxt->addmounts (additional mounts) list to keep order of the requested
+ * flags changes.
+ */
+struct libmnt_addmount *mnt_new_addmount(void)
+{
+	struct libmnt_addmount *ad = calloc(1, sizeof(*ad));
+	if (!ad)
+		return NULL;
+
+	INIT_LIST_HEAD(&ad->mounts);
+	return ad;
+}
+
+void mnt_free_addmount(struct libmnt_addmount *ad)
+{
+	if (!ad)
+		return;
+	list_del(&ad->mounts);
+	free(ad);
+}
+
+static int mnt_context_append_additional_mount(struct libmnt_context *cxt,
+					       struct libmnt_addmount *ad)
+{
+	assert(cxt);
+	assert(ad);
+
+	DBG(CXT, mnt_debug_h(cxt,
+			"mount: add additional flag: 0x%08lx",
+			ad->mountflags));
+
+	list_add_tail(&ad->mounts, &cxt->addmounts);
+	return 0;
+}
+
+static int init_propagation(struct libmnt_context *cxt)
+{
+	char *name;
+	char *opts = (char *) mnt_fs_get_vfs_options(cxt->fs);
+	size_t namesz;
+	struct libmnt_optmap const *maps[1];
+
+	if (!opts)
+		return 0;
+
+	DBG(CXT, mnt_debug_h(cxt, "mount: initialize additional propagation mounts"));
+
+	maps[0] = mnt_get_builtin_optmap(MNT_LINUX_MAP);
+
+	while (!mnt_optstr_next_option(&opts, &name, &namesz, NULL, NULL)) {
+		const struct libmnt_optmap *ent;
+		struct libmnt_addmount *ad;
+		int rc;
+
+		if (!mnt_optmap_get_entry(maps, 1, name, namesz, &ent)
+		    || !ent
+		    || !(ent->id & MS_PROPAGATION))
+			continue;
+
+		ad = mnt_new_addmount();
+		if (!ad)
+			return -ENOMEM;
+
+		ad->mountflags = ent->id;
+		rc = mnt_context_append_additional_mount(cxt, ad);
+		if (rc)
+			return rc;
+
+		cxt->mountflags &= ~ent->id;
+	}
+
+	return 0;
+}
+
 /*
  * this has to be called after mnt_context_evaluate_permissions()
  */
@@ -48,22 +127,15 @@ static int fix_optstr(struct libmnt_context *cxt)
 
 	fs = cxt->fs;
 
-	/* The propagation flags should not be used together with any other
-	 * flags (except MS_REC and MS_SILENT) */
-	if (cxt->mountflags & MS_PROPAGATION)
-		cxt->mountflags &= (MS_PROPAGATION | MS_REC | MS_SILENT);
-
 	/*
 	 * The "user" options is our business (so we can modify the option),
 	 * but exception is command line for /sbin/mount.<type> helpers. Let's
 	 * save the original user=<name> to call the helpers with unchanged
 	 * "user" setting.
-	 *
-	 * Don't check for MNT_MS_USER in cxt->user_mountflags, the flag maybe
-	 * removed by evaluate_permissions().
 	 */
-	if (!mnt_optstr_get_option(fs->user_optstr, "user", &val, &valsz)) {
-		if (val) {
+	if (cxt->user_mountflags & MNT_MS_USER) {
+		if (!mnt_optstr_get_option(fs->user_optstr,
+					"user", &val, &valsz) && val) {
 			cxt->orig_user = strndup(val, valsz);
 			if (!cxt->orig_user) {
 				rc = -ENOMEM;
@@ -97,6 +169,12 @@ static int fix_optstr(struct libmnt_context *cxt)
 		fs->user_optstr = NULL;
 	}
 
+	if (cxt->mountflags & MS_PROPAGATION) {
+		rc = init_propagation(cxt);
+		if (rc)
+			return rc;
+	}
+
 	next = fs->fs_optstr;
 
 #ifdef HAVE_LIBSELINUX
@@ -115,7 +193,7 @@ static int fix_optstr(struct libmnt_context *cxt)
 		 */
 		se_rem = get_linux_version() < KERNEL_VERSION(2, 6, 39);
 	else
-		/* For normal mount we have translate the contexts */
+		/* For normal mount, contexts are translated */
 		se_fix = 1;
 
 	if (!se_rem) {
@@ -157,7 +235,7 @@ static int fix_optstr(struct libmnt_context *cxt)
 			goto done;
 	}
 
-	if (!rc && cxt->user_mountflags & MNT_MS_USER)
+	if (!rc && cxt->restricted && (cxt->user_mountflags & MNT_MS_USER))
 		rc = mnt_optstr_fix_user(&fs->user_optstr);
 
 	/* refresh merged optstr */
@@ -182,7 +260,7 @@ done:
  */
 static int generate_helper_optstr(struct libmnt_context *cxt, char **optstr)
 {
-	struct libmnt_optmap const *maps[1];
+	struct libmnt_optmap const *maps[2];
 	char *next, *name, *val;
 	size_t namesz, valsz;
 	int rc = 0;
@@ -197,6 +275,28 @@ static int generate_helper_optstr(struct libmnt_context *cxt, char **optstr)
 	if (!*optstr)
 		return -ENOMEM;
 
+	if (cxt->user_mountflags & MNT_MS_USER) {
+		/*
+		 * This is unnecessary for real user-mounts as mount.<type>
+		 * helpers have to always follow fstab rather than mount
+		 * options on command line.
+		 *
+		 * But if you call mount.<type> as root then the helper follows
+		 * command line. If there is (for example) "user,exec" in fstab
+		 * then we have to manually append the "exec" back to the options
+		 * string, bacause there is nothing like MS_EXEC (we have only
+		 * MS_NOEXEC in mount flags and we don't care about the original
+		 * mount string in libmount for VFS options).
+		 */
+		if (!(cxt->mountflags & MS_NOEXEC))
+			mnt_optstr_append_option(optstr, "exec", NULL);
+		if (!(cxt->mountflags & MS_NOSUID))
+			mnt_optstr_append_option(optstr, "suid", NULL);
+		if (!(cxt->mountflags & MS_NODEV))
+			mnt_optstr_append_option(optstr, "dev", NULL);
+	}
+
+
 	if (cxt->flags & MNT_FL_SAVED_USER)
 		rc = mnt_optstr_set_option(optstr, "user", cxt->orig_user);
 	if (rc)
@@ -204,12 +304,13 @@ static int generate_helper_optstr(struct libmnt_context *cxt, char **optstr)
 
 	/* remove userspace options with MNT_NOHLPS flag */
 	maps[0] = mnt_get_builtin_optmap(MNT_USERSPACE_MAP);
+	maps[1] = mnt_get_builtin_optmap(MNT_LINUX_MAP);
 	next = *optstr;
 
 	while (!mnt_optstr_next_option(&next, &name, &namesz, &val, &valsz)) {
 		const struct libmnt_optmap *ent;
 
-		mnt_optmap_get_entry(maps, 1, name, namesz, &ent);
+		mnt_optmap_get_entry(maps, 2, name, namesz, &ent);
 		if (ent && ent->id && (ent->mask & MNT_NOHLPS)) {
 			next = name;
 			rc = mnt_optstr_remove_option_at(optstr, name,
@@ -256,8 +357,6 @@ static int evaluate_permissions(struct libmnt_context *cxt)
 		 */
 		cxt->user_mountflags &= ~MNT_MS_OWNER;
 		cxt->user_mountflags &= ~MNT_MS_GROUP;
-		cxt->user_mountflags &= ~MNT_MS_USER;
-		cxt->user_mountflags &= ~MNT_MS_USERS;
 	} else {
 		/*
 		 * user mount
@@ -471,9 +570,46 @@ static int exec_helper(struct libmnt_context *cxt)
 	return rc;
 }
 
+static int do_mount_additional(struct libmnt_context *cxt,
+			       const char *target,
+			       unsigned long flags,
+			       int *syserr)
+{
+	struct list_head *p;
+
+	assert(cxt);
+	assert(target);
+
+	if (syserr)
+		*syserr = 0;
+
+	list_for_each(p, &cxt->addmounts) {
+		int rc;
+		struct libmnt_addmount *ad =
+				list_entry(p, struct libmnt_addmount, mounts);
+
+		DBG(CXT, mnt_debug_h(cxt, "mount(2) changing flag: 0x%08lx %s",
+				ad->mountflags,
+				ad->mountflags & MS_REC ? " (recursive)" : ""));
+
+		rc = mount("none", target, NULL,
+				ad->mountflags | (flags & MS_SILENT), NULL);
+		if (rc) {
+			if (syserr)
+				*syserr = -errno;
+			DBG(CXT, mnt_debug_h(cxt,
+					"mount(2) failed [errno=%d %m]",
+					errno));
+			return rc;
+		}
+	}
+
+	return 0;
+}
+
 /*
  * The default is to use fstype from cxt->fs, this could be overwritten by
- * @try_type argument.
+ * @try_type argument. If @try_type is specified then mount with MS_SILENT.
  *
  * Returns: 0 on success,
  *         >0 in case of mount(2) error (returns syscall errno),
@@ -494,13 +630,23 @@ static int do_mount(struct libmnt_context *cxt, const char *try_type)
 		if (rc)
 			return rc;
 	}
-	if (cxt->helper)
-		return exec_helper(cxt);
 
 	flags = cxt->mountflags;
 	src = mnt_fs_get_srcpath(cxt->fs);
 	target = mnt_fs_get_target(cxt->fs);
 
+	if (cxt->helper) {
+		rc = exec_helper(cxt);
+
+		if (mnt_context_helper_executed(cxt)
+		    && mnt_context_get_helper_status(cxt) == 0
+		    && !list_empty(&cxt->addmounts)
+		    && do_mount_additional(cxt, target, flags, NULL))
+
+			return -MNT_ERR_APPLYFLAGS;
+		return rc;
+	}
+
 	if (!target)
 		return -EINVAL;
 	if (!src) {
@@ -513,6 +659,8 @@ static int do_mount(struct libmnt_context *cxt, const char *try_type)
 
 	if (!(flags & MS_MGC_MSK))
 		flags |= MS_MGC_VAL;
+	if (try_type)
+		flags |= MS_SILENT;
 
 	DBG(CXT, mnt_debug_h(cxt, "%smount(2) "
 			"[source=%s, target=%s, type=%s, "
@@ -521,9 +669,22 @@ static int do_mount(struct libmnt_context *cxt, const char *try_type)
 			src, target, type,
 			flags, cxt->mountdata ? "yes" : "<none>"));
 
-	if (mnt_context_is_fake(cxt))
+	if (mnt_context_is_fake(cxt)) {
+		/*
+		 * fake
+		 */
 		cxt->syscall_status = 0;
-	else {
+
+	} else if (mnt_context_propagation_only(cxt)) {
+		/*
+		 * propagation flags *only*
+		 */
+		if (do_mount_additional(cxt, target, flags, &cxt->syscall_status))
+			return -MNT_ERR_APPLYFLAGS;
+	} else {
+		/*
+		 * regular mount
+		 */
 		if (mount(src, target, type, flags, cxt->mountdata)) {
 			cxt->syscall_status = -errno;
 			DBG(CXT, mnt_debug_h(cxt, "mount(2) failed [errno=%d %m]",
@@ -532,6 +693,16 @@ static int do_mount(struct libmnt_context *cxt, const char *try_type)
 		}
 		DBG(CXT, mnt_debug_h(cxt, "mount(2) success"));
 		cxt->syscall_status = 0;
+
+		/*
+		 * additional mounts for extra propagation flags
+		 */
+		if (!list_empty(&cxt->addmounts)
+		    && do_mount_additional(cxt, target, flags, NULL)) {
+
+			/* TODO: call umount? */
+			return -MNT_ERR_APPLYFLAGS;
+		}
 	}
 
 	if (try_type && cxt->update) {
@@ -587,11 +758,15 @@ static int do_mount_by_pattern(struct libmnt_context *cxt, const char *pattern)
 	if (rc)
 		return rc;
 
+	if (filesystems == NULL)
+		return -MNT_ERR_NOFSTYPE;
+
 	for (fp = filesystems; *fp; fp++) {
 		rc = do_mount(cxt, *fp);
 		if (mnt_context_get_status(cxt))
 			break;
-		if (mnt_context_get_syscall_errno(cxt) != EINVAL)
+		if (mnt_context_get_syscall_errno(cxt) != EINVAL &&
+		    mnt_context_get_syscall_errno(cxt) != ENODEV)
 			break;
 	}
 	mnt_free_filesystems(filesystems);
@@ -626,7 +801,6 @@ int mnt_context_prepare_mount(struct libmnt_context *cxt)
 
 	DBG(CXT, mnt_debug_h(cxt, "mount: preparing"));
 
-	/* TODO: fstab is unnecessary for MS_{MOVE,BIND,STARED,...} */
 	rc = mnt_context_apply_fstab(cxt);
 	if (!rc)
 		rc = mnt_context_merge_mflags(cxt);
@@ -691,9 +865,13 @@ int mnt_context_do_mount(struct libmnt_context *cxt)
 		cxt->mountdata = (char *) mnt_fs_get_fs_options(cxt->fs);
 
 	type = mnt_fs_get_fstype(cxt->fs);
-	if (type)
-		res = do_mount(cxt, NULL);
-	else
+	if (type) {
+		if (strchr(type, ','))
+			/* this only happens if fstab contains list of filesystems */
+			res = do_mount_by_pattern(cxt, type);
+		else
+			res = do_mount(cxt, NULL);
+	} else
 		res = do_mount_by_pattern(cxt, cxt->fstype_pattern);
 
 	if (mnt_context_get_status(cxt)
@@ -718,7 +896,8 @@ int mnt_context_do_mount(struct libmnt_context *cxt)
 		 * system that does not have write support. Check this to avoid
 		 * 'ro' in /proc/mounts and 'rw' in mtab.
 		 */
-		if (!(cxt->mountflags & (MS_RDONLY | MS_PROPAGATION | MS_MOVE))
+		if (!(cxt->mountflags & (MS_RDONLY | MS_MOVE))
+		    && !mnt_context_propagation_only(cxt)
 		    && mnt_is_readonly(mnt_context_get_target(cxt)))
 
 			mnt_context_set_mflags(cxt,
