@@ -161,6 +161,8 @@ static int mnt_parse_mountinfo_line(struct libmnt_fs *fs, char *s)
 		DBG(TAB, mnt_debug("mountinfo parse error: not found separator"));
 		return -EINVAL;
 	}
+	if (p > s + 1)
+		fs->opt_fields = strndup(s + 1, p - s - 1);
 	s = p + 3;
 
 	rc += sscanf(s,	UL_SCNsA" "	/* (8) FS type */
@@ -362,14 +364,14 @@ next_line:
 		if (fgets(buf, sizeof(buf), f) == NULL)
 			return -EINVAL;
 		++*nlines;
-		s = index (buf, '\n');
+		s = strchr (buf, '\n');
 		if (!s) {
 			/* Missing final newline?  Otherwise extremely */
 			/* long line - assume file was corrupted */
 			if (feof(f)) {
 				DBG(TAB, mnt_debug_h(tb,
 					"%s: no final newline",	filename));
-				s = index (buf, '\0');
+				s = strchr (buf, '\0');
 			} else {
 				DBG(TAB, mnt_debug_h(tb,
 					"%s:%d: missing newline at line",
@@ -452,6 +454,43 @@ done:
 	return tid;
 }
 
+static int kernel_fs_postparse(struct libmnt_table *tb,
+			       struct libmnt_fs *fs, pid_t *tid,
+			       const char *filename)
+{
+	int rc = 0;
+	const char *src = mnt_fs_get_srcpath(fs);
+
+	/* This is filesystem description from /proc, so we're in some process
+	 * namespace. Let's remember the process PID.
+	 */
+	if (filename && *tid == -1)
+		*tid = path_to_tid(filename);
+
+	fs->tid = *tid;
+
+	/*
+	 * Convert obscure /dev/root to something more usable
+	 */
+	if (src && strcmp(src, "/dev/root") == 0) {
+		char *spec = mnt_get_kernel_cmdline_option("root=");
+		char *real = NULL;
+
+		DBG(TAB, mnt_debug_h(tb, "root FS: %s", spec));
+		if (spec)
+			real = mnt_resolve_spec(spec, tb->cache);
+		if (real) {
+			DBG(TAB, mnt_debug_h(tb, "canonical root FS: %s", real));
+			rc = mnt_fs_set_source(fs, real);
+			if (!tb->cache)
+				free(real);
+		}
+		free(spec);
+	}
+
+	return rc;
+}
+
 /**
  * mnt_table_parse_stream:
  * @tb: tab pointer
@@ -495,11 +534,9 @@ int mnt_table_parse_stream(struct libmnt_table *tb, FILE *f, const char *filenam
 		if (!rc) {
 			rc = mnt_table_add_fs(tb, fs);
 			fs->flags |= flags;
-			if (tb->fmt == MNT_FMT_MOUNTINFO && filename) {
-				if (tid == -1)
-					tid = path_to_tid(filename);
-				fs->tid = tid;
-			}
+
+			if (rc == 0 && tb->fmt == MNT_FMT_MOUNTINFO)
+				rc = kernel_fs_postparse(tb, fs, &tid, filename);
 		}
 		if (rc) {
 			mnt_free_fs(fs);
@@ -542,7 +579,7 @@ int mnt_table_parse_file(struct libmnt_table *tb, const char *filename)
 	if (!filename || !tb)
 		return -EINVAL;
 
-	f = fopen(filename, "r");
+	f = fopen(filename, "r" UL_CLOEXECSTR);
 	if (f) {
 		rc = mnt_table_parse_stream(tb, f, filename);
 		fclose(f);
@@ -602,7 +639,7 @@ static int __mnt_table_parse_dir(struct libmnt_table *tb, const char *dirname)
 		    !S_ISREG(st.st_mode))
 			continue;
 
-		f = fopen_at(dd, ".", d->d_name, O_RDONLY, "r");
+		f = fopen_at(dd, ".", d->d_name, O_RDONLY|O_CLOEXEC, "r" UL_CLOEXECSTR);
 		if (f) {
 			mnt_table_parse_stream(tb, f, d->d_name);
 			fclose(f);
@@ -642,8 +679,8 @@ static int __mnt_table_parse_dir(struct libmnt_table *tb, const char *dirname)
 		    !S_ISREG(st.st_mode))
 			continue;
 
-		f = fopen_at(dirfd(dir), _PATH_MNTTAB_DIR,
-					d->d_name, O_RDONLY, "r");
+		f = fopen_at(dirfd(dir), _PATH_MNTTAB_DIR, d->d_name,
+				O_RDONLY|O_CLOEXEC, "r" UL_CLOEXECSTR);
 		if (f) {
 			mnt_table_parse_stream(tb, f, d->d_name);
 			fclose(f);
@@ -683,7 +720,6 @@ struct libmnt_table *__mnt_new_table_from_file(const char *filename, int fmt)
 	struct stat st;
 
 	assert(filename);
-
 	if (!filename)
 		return NULL;
 	if (stat(filename, &st))
@@ -712,6 +748,7 @@ struct libmnt_table *__mnt_new_table_from_file(const char *filename, int fmt)
  */
 struct libmnt_table *mnt_new_table_from_file(const char *filename)
 {
+	assert(filename);
 	return __mnt_new_table_from_file(filename, MNT_FMT_GUESS);
 }
 
@@ -726,7 +763,6 @@ struct libmnt_table *mnt_new_table_from_dir(const char *dirname)
 	struct libmnt_table *tb;
 
 	assert(dirname);
-
 	if (!dirname)
 		return NULL;
 	tb = mnt_new_table();
@@ -757,6 +793,8 @@ int mnt_table_set_parser_errcb(struct libmnt_table *tb,
 		int (*cb)(struct libmnt_table *tb, const char *filename, int line))
 {
 	assert(tb);
+	if (!tb)
+		return -EINVAL;
 	tb->errcb = cb;
 	return 0;
 }
@@ -770,8 +808,10 @@ int mnt_table_set_parser_fltrcb(struct libmnt_table *tb,
 		void *data)
 {
 	assert(tb);
+	if (!tb)
+		return -EINVAL;
 
-	DBG(TAB, mnt_debug_h(tb, "set table parser filter"));
+	DBG(TAB, mnt_debug_h(tb, "%s table parser filter", cb ? "set" : "unset"));
 	tb->fltrcb = cb;
 	tb->fltrcb_data = data;
 	return 0;
@@ -823,7 +863,6 @@ int mnt_table_parse_fstab(struct libmnt_table *tb, const char *filename)
 	int rc = 0;
 
 	assert(tb);
-
 	if (!tb)
 		return -EINVAL;
 	if (!filename)
@@ -918,6 +957,9 @@ int mnt_table_parse_mtab(struct libmnt_table *tb, const char *filename)
 {
 	int rc;
 	const char *utab = NULL;
+	struct libmnt_table *u_tb;
+
+	assert(tb);
 
 	if (mnt_has_regular_mtab(&filename, NULL)) {
 
@@ -941,34 +983,33 @@ int mnt_table_parse_mtab(struct libmnt_table *tb, const char *filename)
 		return mnt_table_parse_file(tb, _PATH_PROC_MOUNTS);
 	}
 
+	if (mnt_table_get_nents(tb) == 0)
+		return 0;			/* empty, ignore utab */
 	/*
 	 * try to read user specific information from /run/mount/utabs
 	 */
 	utab = mnt_get_utab_path();
-	if (utab) {
-		struct libmnt_table *u_tb = mnt_new_table();
-		if (u_tb) {
-			u_tb->fmt = MNT_FMT_UTAB;
-			mnt_table_set_parser_fltrcb(u_tb, tb->fltrcb, tb->fltrcb_data);
+	if (!utab || is_file_empty(utab))
+		return 0;
 
-			if (mnt_table_parse_file(u_tb, utab) != 0) {
-				mnt_free_table(u_tb);
-				u_tb = NULL;
-			}
-		}
+	u_tb = mnt_new_table();
+	if (!u_tb)
+		return -ENOMEM;
 
-		if (u_tb) {
-			struct libmnt_fs *u_fs;
-			struct libmnt_iter itr;
+	u_tb->fmt = MNT_FMT_UTAB;
+	mnt_table_set_parser_fltrcb(u_tb, tb->fltrcb, tb->fltrcb_data);
 
-			mnt_reset_iter(&itr, MNT_ITER_BACKWARD);
+	if (mnt_table_parse_file(u_tb, utab) == 0) {
+		struct libmnt_fs *u_fs;
+		struct libmnt_iter itr;
 
-			/*  merge user options into mountinfo from kernel */
-			while(mnt_table_next_fs(u_tb, &itr, &u_fs) == 0)
-				mnt_table_merge_user_fs(tb, u_fs);
+		mnt_reset_iter(&itr, MNT_ITER_BACKWARD);
 
-			mnt_free_table(u_tb);
-		}
+		/*  merge user options into mountinfo from kernel */
+		while(mnt_table_next_fs(u_tb, &itr, &u_fs) == 0)
+			mnt_table_merge_user_fs(tb, u_fs);
 	}
+
+	mnt_free_table(u_tb);
 	return 0;
 }
