@@ -87,6 +87,8 @@ int mnt_update_set_filename(struct libmnt_update *upd, const char *filename,
 	int rw = 0;
 
 	assert(upd);
+	if (!upd)
+		return -EINVAL;
 
 	/* filename explicitly defined */
 	if (filename) {
@@ -129,6 +131,7 @@ int mnt_update_set_filename(struct libmnt_update *upd, const char *filename,
  */
 const char *mnt_update_get_filename(struct libmnt_update *upd)
 {
+	assert(upd);
 	return upd ? upd->filename : NULL;
 }
 
@@ -141,6 +144,7 @@ const char *mnt_update_get_filename(struct libmnt_update *upd)
  */
 int mnt_update_is_ready(struct libmnt_update *upd)
 {
+	assert(upd);
 	return upd ? upd->ready : FALSE;
 }
 
@@ -169,7 +173,7 @@ int mnt_update_set_fs(struct libmnt_update *upd, unsigned long mountflags,
 		return -EINVAL;
 
 	DBG(UPDATE, mnt_debug_h(upd,
-			"reseting FS [fs=0x%p, target=%s, flags=0x%08lx]",
+			"resetting FS [fs=0x%p, target=%s, flags=0x%08lx]",
 			fs, target, mountflags));
 	if (fs) {
 		DBG(UPDATE, mnt_debug_h(upd, "FS template:"));
@@ -200,7 +204,7 @@ int mnt_update_set_fs(struct libmnt_update *upd, unsigned long mountflags,
 
 	} else if (fs) {
 		if (upd->userspace_only && !(mountflags & MS_MOVE)) {
-			int rc = utab_new_entry(upd, fs, mountflags);
+			rc = utab_new_entry(upd, fs, mountflags);
 			if (rc)
 				return rc;
 		} else {
@@ -225,6 +229,7 @@ int mnt_update_set_fs(struct libmnt_update *upd, unsigned long mountflags,
  */
 struct libmnt_fs *mnt_update_get_fs(struct libmnt_update *upd)
 {
+	assert(upd);
 	return upd ? upd->fs : NULL;
 }
 
@@ -236,6 +241,7 @@ struct libmnt_fs *mnt_update_get_fs(struct libmnt_update *upd)
  */
 unsigned long mnt_update_get_mflags(struct libmnt_update *upd)
 {
+	assert(upd);
 	return upd ? upd->mountflags : 0;
 }
 
@@ -250,6 +256,7 @@ int mnt_update_force_rdonly(struct libmnt_update *upd, int rdonly)
 {
 	int rc = 0;
 
+	assert(upd);
 	if (!upd || !upd->fs)
 		return -EINVAL;
 
@@ -503,6 +510,7 @@ static int update_table(struct libmnt_update *upd, struct libmnt_table *tb)
 	int rc, fd;
 	char *uq = NULL;
 
+	assert(upd);
 	if (!tb || !upd->filename)
 		return -EINVAL;
 
@@ -512,12 +520,11 @@ static int update_table(struct libmnt_update *upd, struct libmnt_table *tb)
 	if (fd < 0)
 		return fd;	/* error */
 
-	f = fdopen(fd, "w");
+	f = fdopen(fd, "w" UL_CLOEXECSTR);
 	if (f) {
 		struct stat st;
 		struct libmnt_iter itr;
 		struct libmnt_fs *fs;
-		int fd;
 
 		mnt_reset_iter(&itr, MNT_ITER_FORWARD);
 		while(mnt_table_next_fs(tb, &itr, &fs) == 0) {
@@ -538,7 +545,6 @@ static int update_table(struct libmnt_update *upd, struct libmnt_table *tb)
 			goto leave;
 		}
 
-		fd = fileno(f);
 		rc = fchmod(fd, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH) ? -errno : 0;
 
 		if (!rc && stat(upd->filename, &st) == 0)
@@ -546,18 +552,38 @@ static int update_table(struct libmnt_update *upd, struct libmnt_table *tb)
 			rc = fchown(fd, st.st_uid, st.st_gid) ? -errno : 0;
 
 		fclose(f);
-		rc = rename(uq, upd->filename) ? -errno : 0;
+		f = NULL;
+
+		if (!rc)
+			rc = rename(uq, upd->filename) ? -errno : 0;
 	} else {
 		rc = -errno;
 		close(fd);
 	}
 
 leave:
+	if (f)
+		fclose(f);
+
 	unlink(uq);	/* be paranoid */
 	free(uq);
 	return rc;
 }
 
+static int add_file_entry(struct libmnt_table *tb, struct libmnt_update *upd)
+{
+	struct libmnt_fs *fs;
+
+	assert(upd);
+
+	fs = mnt_copy_fs(NULL, upd->fs);
+	if (!fs)
+		return -ENOMEM;
+
+	mnt_table_add_fs(tb, fs);
+	return update_table(upd, tb);
+}
+
 static int update_add_entry(struct libmnt_update *upd, struct libmnt_lock *lc)
 {
 	struct libmnt_table *tb;
@@ -575,16 +601,8 @@ static int update_add_entry(struct libmnt_update *upd, struct libmnt_lock *lc)
 
 	tb = __mnt_new_table_from_file(upd->filename,
 			upd->userspace_only ? MNT_FMT_UTAB : MNT_FMT_MTAB);
-	if (tb) {
-		struct libmnt_fs *fs = mnt_copy_fs(NULL, upd->fs);
-		if (!fs)
-			rc = -ENOMEM;
-		else {
-			mnt_table_add_fs(tb, fs);
-			rc = update_table(upd, tb);
-		}
-	}
-
+	if (tb)
+		rc = add_file_entry(tb, upd);
 	if (lc)
 		mnt_unlock_file(lc);
 
@@ -617,7 +635,6 @@ static int update_remove_entry(struct libmnt_update *upd, struct libmnt_lock *lc
 			mnt_free_fs(rem);
 		}
 	}
-
 	if (lc)
 		mnt_unlock_file(lc);
 
@@ -630,6 +647,7 @@ static int update_modify_target(struct libmnt_update *upd, struct libmnt_lock *l
 	struct libmnt_table *tb = NULL;
 	int rc = 0;
 
+	assert(upd);
 	DBG(UPDATE, mnt_debug_h(upd, "%s: modify target", upd->filename));
 
 	if (lc)
@@ -687,7 +705,8 @@ static int update_modify_options(struct libmnt_update *upd, struct libmnt_lock *
 				rc = mnt_fs_set_options(cur, mnt_fs_get_options(fs));
 			if (!rc)
 				rc = update_table(upd, tb);
-		}
+		} else
+			rc = add_file_entry(tb, upd);	/* not found, add new */
 	}
 
 	if (lc)
@@ -717,8 +736,7 @@ int mnt_update_table(struct libmnt_update *upd, struct libmnt_lock *lc)
 	int rc = -EINVAL;
 
 	assert(upd);
-
-	if (!upd->filename || !upd)
+	if (!upd || !upd->filename)
 		return -EINVAL;
 	if (!upd->ready)
 		return 0;
@@ -780,6 +798,7 @@ static int update(const char *target, struct libmnt_fs *fs, unsigned long mountf
 
 	rc = mnt_update_table(upd, NULL);
 done:
+	mnt_free_update(upd);
 	return rc;
 }
 
