@@ -356,7 +356,7 @@ static void init_tty(struct login_context *cxt)
 
 	cxt->tty_mode = (mode_t) getlogindefs_num("TTYPERM", TTY_MODE);
 
-	get_terminal_name(&cxt->tty_path, &cxt->tty_name, &cxt->tty_number);
+	get_terminal_name(0, &cxt->tty_path, &cxt->tty_name, &cxt->tty_number);
 
 	/*
 	 * In case login is suid it was possible to use a hardlink as stdin
@@ -395,7 +395,7 @@ static void init_tty(struct login_context *cxt)
 	}
 
 	/* Kill processes left on this tty */
-	tcsetattr(0, TCSAFLUSH, &ttt);
+	tcsetattr(0, TCSANOW, &ttt);
 
 	/*
 	 * Let's close file decriptors before vhangup
@@ -1312,10 +1312,10 @@ int main(int argc, char **argv)
 		close(cnt);
 
 	setpgrp();	 /* set pgid to pid this means that setsid() will fail */
+	init_tty(&cxt);
 
 	openlog("login", LOG_ODELAY, LOG_AUTHPRIV);
 
-	init_tty(&cxt);
 	init_loginpam(&cxt);
 
 	/* login -f, then the user has already been authenticated */
