@@ -36,7 +36,7 @@
 
 /* 1999-02-01	Jean-Francois Bignolles: added option '-m' to display
  *		monday as the first day of the week.
- * 1999-02-22 Arkadiusz Mikiewicz <misiek@pld.ORG.PL>
+ * 1999-02-22 Arkadiusz Miśkiewicz <misiek@pld.ORG.PL>
  * - added Native Language Support
  *
  * 2000-09-01  Michael Charles Pruznick <dummy@netwiz.net>
@@ -58,17 +58,19 @@
 #include <sys/types.h>
 
 #include <ctype.h>
+#include <getopt.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <time.h>
 #include <unistd.h>
-#include <err.h>
 #include <errno.h>
 
 #include "c.h"
+#include "closestream.h"
 #include "nls.h"
 #include "mbsalign.h"
+#include "strutils.h"
 
 #if defined(HAVE_LIBNCURSES) || defined(HAVE_LIBNCURSESW)
 
@@ -91,12 +93,12 @@ my_putstring(char *s) {
 }
 
 static const char *
-my_tgetstr(char *s, char *ss) {
+my_tgetstr(char *s __attribute__ ((__unused__)), char *ss) {
     const char* ret = tigetstr(ss);
     if (!ret || ret==(char*)-1)
-        return "";
+	return "";
     else
-        return ret;
+	return ret;
 }
 
 #elif defined(HAVE_LIBTERMCAP)
@@ -118,12 +120,12 @@ my_putstring(char *s) {
 }
 
 static const char *
-my_tgetstr(char *s, char *ss) {
+my_tgetstr(char *s, char *ss __attribute__ ((__unused__))) {
     const char* ret = tgetstr(s, &strbuf);
     if (!ret)
-        return "";
+	return "";
     else
-        return ret;
+	return ret;
 }
 
 #else /* ! (HAVE_LIBTERMCAP || HAVE_LIBNCURSES || HAVE_LIBNCURSESW) */
@@ -252,7 +254,7 @@ void do_monthly(int, int, int, struct fmt_st*);
 void monthly(int, int, int);
 void monthly3(int, int, int);
 void trim_trailing_spaces(char *);
-void usage(void);
+static void __attribute__ ((__noreturn__)) usage(FILE * out);
 void headers_init(void);
 
 int
@@ -262,9 +264,22 @@ main(int argc, char **argv) {
 	int ch, day, month, year, yflag;
 	int num_months = NUM_MONTHS;
 
+	static const struct option longopts[] = {
+		{"one", no_argument, NULL, '1'},
+		{"three", no_argument, NULL, '3'},
+		{"sunday", no_argument, NULL, 's'},
+		{"monday", no_argument, NULL, 'm'},
+		{"julian", no_argument, NULL, 'j'},
+		{"year", no_argument, NULL, 'y'},
+		{"version", no_argument, NULL, 'V'},
+		{"help", no_argument, NULL, 'h'},
+		{NULL, 0, NULL, 0}
+	};
+
 	setlocale(LC_ALL, "");
 	bindtextdomain(PACKAGE, LOCALEDIR);
 	textdomain(PACKAGE);
+	atexit(close_stdout);
 
 #if defined(HAVE_LIBNCURSES) || defined(HAVE_LIBNCURSESW) || defined(HAVE_LIBTERMCAP)
 	if ((term = getenv("TERM"))) {
@@ -290,17 +305,17 @@ main(int argc, char **argv) {
 	 * This is needed to support first_weekday=2 and first_workday=1 for
 	 * the rare case where working days span across 2 weeks.
 	 * This shell script shows the combinations and calculations involved:
-
-	 for LANG in en_US ru_RU fr_FR csb_PL POSIX; do
-	   printf "%s:\t%s + %s -1 = " $LANG $(locale week-1stday first_weekday)
-	   date -d"$(locale week-1stday) +$(($(locale first_weekday)-1))day" +%w
-	 done
-
-	 en_US:  19971130 + 1 -1 = 0  #0 = sunday
-	 ru_RU:  19971130 + 2 -1 = 1
-	 fr_FR:  19971201 + 1 -1 = 1
-	 csb_PL: 19971201 + 2 -1 = 2
-	 POSIX:  19971201 + 7 -1 = 0
+	 *
+	 * for LANG in en_US ru_RU fr_FR csb_PL POSIX; do
+	 *   printf "%s:\t%s + %s -1 = " $LANG $(locale week-1stday first_weekday)
+	 *   date -d"$(locale week-1stday) +$(($(locale first_weekday)-1))day" +%w
+	 * done
+	 *
+	 * en_US:  19971130 + 1 -1 = 0  #0 = sunday
+	 * ru_RU:  19971130 + 2 -1 = 1
+	 * fr_FR:  19971201 + 1 -1 = 1
+	 * csb_PL: 19971201 + 2 -1 = 2
+	 * POSIX:  19971201 + 7 -1 = 0
 	 */
 	{
 		int wfd;
@@ -314,7 +329,7 @@ main(int argc, char **argv) {
 #endif
 
 	yflag = 0;
-	while ((ch = getopt(argc, argv, "13mjsyV")) != -1)
+	while ((ch = getopt_long(argc, argv, "13mjsyVh", longopts, NULL)) != -1)
 		switch(ch) {
 		case '1':
 			num_months = 1;		/* default */
@@ -335,12 +350,13 @@ main(int argc, char **argv) {
 			yflag = 1;
 			break;
 		case 'V':
-			printf(_("%s from %s\n"), program_invocation_short_name,
-			       PACKAGE_STRING);
+			printf(UTIL_LINUX_VERSION);
 			return EXIT_SUCCESS;
+		case 'h':
+			usage(stdout);
 		case '?':
 		default:
-			usage();
+			usage(stderr);
 		}
 	argc -= optind;
 	argv += optind;
@@ -351,20 +367,23 @@ main(int argc, char **argv) {
 	day = month = year = 0;
 	switch(argc) {
 	case 3:
-		if ((day = atoi(*argv++)) < 1 || day > 31)
-			errx(1, _("illegal day value: use 1-%d"), 31);
+		day = strtos32_or_err(*argv++, _("illegal day value"));
+                if (day < 1 || 31 < day)
+			errx(EXIT_FAILURE, _("illegal day value: use 1-%d"), 31);
 		/* FALLTHROUGH */
 	case 2:
-		if ((month = atoi(*argv++)) < 1 || month > 12)
-			errx(1, _("illegal month value: use 1-12"));
+		month = strtos32_or_err(*argv++, _("illegal month value: use 1-12"));
+		if (month < 1 || 12 < month)
+			errx(EXIT_FAILURE, _("illegal month value: use 1-12"));
 		/* FALLTHROUGH */
 	case 1:
-		if ((year = atoi(*argv)) < 1 || year > 9999)
-			errx(1, _("illegal year value: use 1-9999"));
+		year = strtos32_or_err(*argv++, _("illegal year value: use 1-9999"));
+		if (year < 1 || 9999 < year)
+			errx(EXIT_FAILURE, _("illegal year value: use 1-9999"));
 		if (day) {
 			int dm = days_in_month[leap_year(year)][month];
 			if (day > dm)
-				errx(1, _("illegal day value: use 1-%d"), dm);
+				errx(EXIT_FAILURE, _("illegal day value: use 1-%d"), dm);
 			day = day_in_year(day, month, year);
 		} else if ((local_time->tm_year + 1900) == year) {
 			day = local_time->tm_yday + 1;
@@ -378,7 +397,7 @@ main(int argc, char **argv) {
 		month = local_time->tm_mon + 1;
 		break;
 	default:
-		usage();
+		usage(stderr);
 	}
 	headers_init();
 
@@ -399,33 +418,39 @@ main(int argc, char **argv) {
 
 void headers_init(void)
 {
-  int i, wd;
-  char *cur_dh = day_headings, *cur_j_dh = j_day_headings;
+	int i, wd;
+	char *cur_dh = day_headings, *cur_j_dh = j_day_headings;
 
-  strcpy(day_headings,"");
-  strcpy(j_day_headings,"");
+	strcpy(day_headings, "");
+	strcpy(j_day_headings, "");
 
-  for(i = 0 ; i < 7 ; i++ ) {
-     ssize_t space_left;
-     wd = (i + weekstart) % 7;
+	for (i = 0; i < 7; i++) {
+		ssize_t space_left;
+		wd = (i + weekstart) % 7;
 
-     if (i)
-        strcat(cur_dh++, " ");
-     space_left = sizeof(day_headings) - (cur_dh - day_headings);
-     if(space_left <= 2)
-        break;
-     cur_dh += center_str(nl_langinfo(ABDAY_1+wd), cur_dh, space_left, 2);
+		if (i)
+			strcat(cur_dh++, " ");
+		space_left =
+		    sizeof(day_headings) - (cur_dh - day_headings);
+		if (space_left <= 2)
+			break;
+		cur_dh +=
+		    center_str(nl_langinfo(ABDAY_1 + wd), cur_dh,
+			       space_left, 2);
 
-     if (i)
-        strcat(cur_j_dh++, " ");
-     space_left = sizeof(j_day_headings) - (cur_j_dh - j_day_headings);
-     if(space_left <= 3)
-        break;
-     cur_j_dh += center_str(nl_langinfo(ABDAY_1+wd), cur_j_dh, space_left, 3);
-  }
+		if (i)
+			strcat(cur_j_dh++, " ");
+		space_left =
+		    sizeof(j_day_headings) - (cur_j_dh - j_day_headings);
+		if (space_left <= 3)
+			break;
+		cur_j_dh +=
+		    center_str(nl_langinfo(ABDAY_1 + wd), cur_j_dh,
+			       space_left, 3);
+	}
 
-  for (i = 0; i < 12; i++)
-     full_month[i] = nl_langinfo(MON_1+i);
+	for (i = 0; i < 12; i++)
+		full_month[i] = nl_langinfo(MON_1 + i);
 }
 
 void
@@ -472,7 +497,7 @@ monthly(int day, int month, int year) {
 	do_monthly(day, month, year, &out);
 	for (i = 0; i < FMT_ST_LINES; i++) {
 		my_putstring(out.s[i]);
-		putchar('\n');
+		my_putstring("\n");
 	}
 }
 
@@ -506,18 +531,21 @@ monthly3(int day, int month, int year) {
 	do_monthly(day, month,      year,      &out_curm);
 	do_monthly(day, next_month, next_year, &out_next);
 
-        width = (julian ? J_WEEK_LEN : WEEK_LEN) -1;
-	for (i = 0; i < 2; i++)
-		printf("%s  %s  %s\n", out_prev.s[i], out_curm.s[i], out_next.s[i]);
+	width = (julian ? J_WEEK_LEN : WEEK_LEN) -1;
+	for (i = 0; i < 2; i++) {
+		snprintf(lineout, sizeof(lineout),
+			"%s  %s  %s\n", out_prev.s[i], out_curm.s[i], out_next.s[i]);
+		my_putstring(lineout);
+	}
 	for (i = 2; i < FMT_ST_LINES; i++) {
 		int w1, w2, w3;
 		w1 = w2 = w3 = width;
 
 #if defined(HAVE_LIBNCURSES) || defined(HAVE_LIBNCURSESW) || defined(HAVE_LIBTERMCAP)
-                /* adjust width to allow for non printable characters */
-                w1 += (out_prev.s[i] == Hrow ? Slen : 0);
-                w2 += (out_curm.s[i] == Hrow ? Slen : 0);
-                w3 += (out_next.s[i] == Hrow ? Slen : 0);
+		/* adjust width to allow for non printable characters */
+		w1 += (out_prev.s[i] == Hrow ? Slen : 0);
+		w2 += (out_curm.s[i] == Hrow ? Slen : 0);
+		w3 += (out_next.s[i] == Hrow ? Slen : 0);
 #endif
 		snprintf(lineout, sizeof(lineout), "%-*s  %-*s  %-*s\n",
 		       w1, out_prev.s[i],
@@ -536,7 +564,7 @@ j_yearly(int day, int year) {
 
 	snprintf(lineout, sizeof(lineout), "%d", year);
 	center(lineout, J_WEEK_LEN*2 + J_HEAD_SEP - 1, 0);
-	printf("\n\n");
+	my_putstring("\n\n");
 
 	for (i = 0; i < 12; i++)
 		day_array(day, i + 1, year, days[i]);
@@ -545,8 +573,10 @@ j_yearly(int day, int year) {
 	for (month = 0; month < 12; month += 2) {
 		center(full_month[month], J_WEEK_LEN-1, J_HEAD_SEP+1);
 		center(full_month[month + 1], J_WEEK_LEN-1, 0);
-		printf("\n%s%*s %s\n", j_day_headings, J_HEAD_SEP, "",
-		    j_day_headings);
+		snprintf(lineout, sizeof(lineout),
+			    "\n%s%*s %s\n", j_day_headings, J_HEAD_SEP, "",
+			    j_day_headings);
+		my_putstring(lineout);
 		for (row = 0; row < 6; row++) {
 			p = lineout;
 			for (which_cal = 0; which_cal < 2; which_cal++) {
@@ -558,21 +588,24 @@ j_yearly(int day, int year) {
 			*p = '\0';
 			trim_trailing_spaces(lineout);
 			my_putstring(lineout);
-			putchar('\n');
+			my_putstring("\n");
 		}
 	}
-	printf("\n");
+	my_putstring("\n");
 }
 
 void
 yearly(int day, int year) {
 	int col, *dp, i, month, row, which_cal;
 	int days[12][MAXDAYS];
-	char *p, lineout[100];
+	char *p;
+	char lineout[ sizeof(day_headings) + 2 +
+		      sizeof(day_headings) + 2 +
+		      sizeof(day_headings) + 1 ];
 
 	snprintf(lineout, sizeof(lineout), "%d", year);
 	center(lineout, WEEK_LEN*3 + HEAD_SEP*2 - 1, 0);
-	printf("\n\n");
+	my_putstring("\n\n");
 
 	for (i = 0; i < 12; i++)
 		day_array(day, i + 1, year, days[i]);
@@ -582,8 +615,10 @@ yearly(int day, int year) {
 		center(full_month[month], WEEK_LEN-1, HEAD_SEP+1);
 		center(full_month[month + 1], WEEK_LEN-1, HEAD_SEP+1);
 		center(full_month[month + 2], WEEK_LEN-1, 0);
-		printf("\n%s%*s %s%*s %s\n", day_headings, HEAD_SEP,
-		    "", day_headings, HEAD_SEP, "", day_headings);
+		snprintf(lineout, sizeof(lineout),
+			"\n%s%*s %s%*s %s\n", day_headings, HEAD_SEP,
+			"", day_headings, HEAD_SEP, "", day_headings);
+		my_putstring(lineout);
 		for (row = 0; row < 6; row++) {
 			p = lineout;
 			for (which_cal = 0; which_cal < 3; which_cal++) {
@@ -595,10 +630,10 @@ yearly(int day, int year) {
 			*p = '\0';
 			trim_trailing_spaces(lineout);
 			my_putstring(lineout);
-			putchar('\n');
+			my_putstring("\n");
 		}
 	}
-	putchar('\n');
+	my_putstring("\n");
 }
 
 /*
@@ -718,8 +753,7 @@ ascii_day(char *p, int day) {
 }
 
 void
-trim_trailing_spaces(s)
-	char *s;
+trim_trailing_spaces(char *s)
 {
 	char *p;
 
@@ -745,23 +779,35 @@ center_str(const char* src, char* dest, size_t dest_size, size_t width)
 }
 
 void
-center(str, len, separate)
-	const char *str;
-	size_t len;
-	int separate;
+center(const char *str, size_t len, int separate)
 {
 	char lineout[FMT_ST_CHARS];
+
 	center_str(str, lineout, ARRAY_SIZE(lineout), len);
-	fputs(lineout, stdout);
-	if (separate)
-		printf("%*s", separate, "");
+	my_putstring(lineout);
+
+	if (separate) {
+		snprintf(lineout, sizeof(lineout), "%*s", separate, "");
+		my_putstring(lineout);
+	}
 }
 
-void
-usage()
+static void __attribute__ ((__noreturn__)) usage(FILE * out)
 {
-
-	fprintf(stderr, _("usage: %s [-13smjyV] [[[day] month] year]\n"),
+	fputs(_("\nUsage:\n"), out);
+	fprintf(out,
+	      _(" %s [options] [[[day] month] year]\n"),
 			program_invocation_short_name);
-	exit(EXIT_FAILURE);
+
+	fputs(_("\nOptions:\n"), out);
+	fputs(_(" -1, --one        show only current month (default)\n"
+	        " -3, --three      show previous, current and next month\n"
+		" -s, --sunday     Sunday as first day of week\n"
+		" -m, --monday     Monday as first day of week\n"
+		" -j, --julian     output Julian dates\n"
+		" -y, --year       show whole current year\n"
+		" -V, --version    display version information and exit\n"
+		" -h, --help       display this help text and exit\n\n"), out);
+
+	exit(out == stderr ? EXIT_FAILURE : EXIT_SUCCESS);
 }
