@@ -398,6 +398,9 @@ static void close_tray(int fd)
 static int eject_cdrom(int fd)
 {
 #if defined(CDROMEJECT)
+	int ret = ioctl(fd, CDROM_LOCKDOOR, 0);
+	if (ret < 0)
+		return 0;
 	return ioctl(fd, CDROMEJECT) >= 0;
 #elif defined(CDIOCEJECT)
 	return ioctl(fd, CDIOCEJECT) >= 0;
@@ -499,7 +502,7 @@ static int read_speed(const char *devname)
 	if (!f)
 		err(EXIT_FAILURE, _("cannot open %s"), _PATH_PROC_CDROMINFO);
 
-	name = rindex(devname, '/') + 1;
+	name = strrchr(devname, '/') + 1;
 
 	while (name && !feof(f)) {
 		char line[512];
@@ -655,7 +658,7 @@ static void umount_one(const char *name)
 			err(EXIT_FAILURE, _("cannot set group id"));
 
 		if (setuid(getuid()) < 0)
-			err(EXIT_FAILURE, _("eject: cannot set user id"));
+			err(EXIT_FAILURE, _("cannot set user id"));
 
 		if (p_option)
 			execl("/bin/umount", "/bin/umount", name, "-n", NULL);
@@ -837,7 +840,7 @@ static char *get_subsystem(char *chain, char *buf, size_t bufsz)
 		memcpy(chain + len, SUBSYSTEM_LINKNAME, sizeof(SUBSYSTEM_LINKNAME));
 
 		/* try if subsystem symlink exists */
-		sz = readlink(chain, buf, bufsz);
+		sz = readlink(chain, buf, bufsz - 1);
 
 		/* remove last subsystem from chain */
 		chain[len] = '\0';
